#include <Arduino.h>
#include <TM1637Display.h>
#include <Ticker.h>
#include <DNSServer.h>
#include <WiFi.h>
#include <AsyncTCP.h>
#include <ESPmDNS.h>
#include <ArduinoOTA.h>
#include <ESPAsyncWebServer.h>
#include <LittleFS.h>
#include <Wire.h>
#include <Adafruit_INA219.h>
#include <ArduinoJson.h>

// local includes
#include "defaults.h"
#include "webui.h"
#include "config.h"
#include "globals.h"
#include "dtc.h"
#include "common.h"
#include "lora_net.h"
#include "oled_display.h"

#ifdef WIFI_CLIENT
#include <WiFiMulti.h>

const char *ssid = QUOTE(WIFI_SSID);
const char *password = QUOTE(WIFI_PASSWORD);
const uint32_t connectTimeoutMs = 5000;

WiFiMulti wifiMulti;
#endif

void SevenSeg_Output();
void FactionTicker_callback();
void inputGetterTicker_callback();
void powerMonitorTicker_callback();
void EEPROMCyclicPDS_callback();
void toggleWiFiAP(boolean shutdown = false);
void SystemShutdown();
void SetBatteryType(batteryType_t type);
void ProcessKeyCombos();
void OverrideDisplay(const uint8_t *message, uint32_t time);
void LoRaPublish_callback();

#ifdef WIFI_CLIENT
void wifiMaintainConnectionTicker_callback();
Ticker WiFiMaintainConnectionTicker(wifiMaintainConnectionTicker_callback, 1000, 0, MILLIS);
#endif

uint32_t getESPChipID();

TM1637Display disp_FAC_1(CLK, DIO_FAC_1_7SEG);
TM1637Display disp_FAC_2(CLK, DIO_FAC_2_7SEG);
TM1637Display disp_FAC_3(CLK, DIO_FAC_3_7SEG);

Adafruit_INA219 ina219;

#ifdef CAPTIVE
DNSServer dnsServer;
#endif
AsyncWebServer server(80);

Ticker FactionTicker(FactionTicker_callback, 1000, 0, MILLIS);
Ticker InputGetterTicker(inputGetterTicker_callback, 250, 0, MILLIS);
Ticker PowerMonitorTicker(powerMonitorTicker_callback, 5000, 0, MILLIS);
Ticker EEPROMCyclicPDSTicker(EEPROMCyclicPDS_callback, 60000, 0, MILLIS);
Ticker LoRaPublishTicker(LoRaPublish_callback, 60000, 0, MILLIS);

uint8_t Faction_1_dot = 0;
uint8_t Faction_2_dot = 0;
uint8_t Faction_3_dot = 0;

uint32_t DisplayOverrideFlag = 0;
uint8_t *DisplayOverrideValue = 0;

Globals_t globals;

const uint8_t sevenSeg_bat[] = {0x00, 0b01111100, 0b01110111, 0b01111000};
const uint8_t sevenSeg_low[] = {0b00111000, 0b01011100, 0x00, 0x00};
const uint8_t sevenSeg_net[] = {0b01010100, 0b01111001, 0b01111000, 0x00};
const uint8_t sevenSeg_ota[] = {0x3F, 0x78, 0x77, 0x00};
const uint8_t sevenSeg_flsh[] = {0x71, 0x38, 0x6D, 0x76};
const uint8_t sevenSeg_file[] = {0x71, 0x30, 0x38, 0x79};

void setup()
{
  setCpuFrequencyMhz(80);
  WiFi.setAutoReconnect(false);
  WiFi.persistent(false);
  WiFi.disconnect();

  Serial.begin(115200);
  Serial.print("\n\n\n");

  strcpy(globals.DeviceName, DEVICE_NAME);
  snprintf(globals.DeviceName_ID, 42, "%s_%08X", globals.DeviceName, getESPChipID());

  if (LoRa_Init())
    LoRaPublishTicker.start();
  OLED_Init();

  pinMode(DIO_FAC_1_TRG, INPUT_PULLUP);
  pinMode(DIO_FAC_2_TRG, INPUT_PULLUP);
  pinMode(DIO_FAC_3_TRG, INPUT_PULLUP);

#ifdef SERIAL_DEBUG
  Serial.setDebugOutput(true);
#endif

  if (ina219.begin())
    PowerMonitorTicker.start();
  else
    Serial.println("Failed to find INA219 chip");

  LittleFS.begin();

#ifdef WIFI_CLIENT
  WiFi.mode(WIFI_STA);
  WiFi.setHostname(globals.DeviceName_ID);
  wifiMulti.addAP(QUOTE(WIFI_SSID), QUOTE(WIFI_PASSWORD));
  WiFiMaintainConnectionTicker.start();
#else
  WiFi.mode(WIFI_AP);
  WiFi.begin(QUOTE(DEVICE_NAME), QUOTE(WIFI_AP_PASSWORD));
  WiFi.setSleep(true);
  WiFi.mode(WIFI_OFF);
#endif

  InitEEPROM();

  ArduinoOTA.setPort(8266);
  ArduinoOTA.setHostname(globals.DeviceName_ID);
  ArduinoOTA.setPassword(QUOTE(ADMIN_PASSWORD));

  ArduinoOTA.onStart([]()
                     {
                       disp_FAC_1.setBrightness(7);
                       disp_FAC_2.setBrightness(7);
                       disp_FAC_3.setBrightness(7);
                       disp_FAC_1.setSegments(sevenSeg_ota);
                       disp_FAC_3.clear();
                       
                       if (ArduinoOTA.getCommand() == U_FLASH)
                       {
                         disp_FAC_2.setSegments(sevenSeg_flsh);
                       }
                       else
                       {
                         disp_FAC_2.setSegments(sevenSeg_file);
                         LittleFS.end();
                       } });

  ArduinoOTA.onEnd([]()
                   { 
                       const uint8_t seg_done[] = {0x5E, 0x3F, 0x54, 0x79};
                       disp_FAC_1.setSegments(seg_done);
                       disp_FAC_2.clear();
                       disp_FAC_3.clear(); });

  ArduinoOTA.onProgress([](unsigned int progress, unsigned int total)
                        { disp_FAC_3.showNumberDecEx((progress / (total / 100))); });

  ArduinoOTA.onError([](ota_error_t error)
                     {
                       Serial.printf("Error[%u]: ", error);
                       if (error == OTA_AUTH_ERROR)
                         Serial.println("Auth Failed");
                       else if (error == OTA_BEGIN_ERROR)
                         Serial.println("Begin Failed");
                       else if (error == OTA_CONNECT_ERROR)
                         Serial.println("Connect Failed");
                       else if (error == OTA_RECEIVE_ERROR)
                         Serial.println("Receive Failed");
                       else if (error == OTA_END_ERROR)
                         Serial.println("End Failed"); });
  ArduinoOTA.begin();

#ifdef CAPTIVE
  dnsServer.start(53, "*", WiFi.softAPIP());
#endif

  initWebUI();

  EEPROMCyclicPDSTicker.start();
  FactionTicker.start();
  InputGetterTicker.start();
  Serial.println("Setup Done");
}

void loop()
{
  EEPROMCyclicPDSTicker.update();
  FactionTicker.update();
  InputGetterTicker.update();
  PowerMonitorTicker.update();
  LoRaPublishTicker.update();
  ArduinoOTA.handle();
  SevenSeg_Output();
  EEPROM_Process();
  OLED_Process();

#ifdef CAPTIVE
  dnsServer.processNextRequest();
#endif
#ifdef WIFI_CLIENT
  WiFiMaintainConnectionTicker.update();
#endif
  if (globals.systemStatus == sysStat_Shutdown)
    SystemShutdown();
  yield();
}

String macToString(const unsigned char *mac)
{
  char buf[20];
  snprintf(buf, sizeof(buf), "%02x:%02x:%02x:%02x:%02x:%02x",
           mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
  return String(buf);
}

void SevenSeg_Output()
{
  if (DisplayOverrideFlag > millis())
  {
    disp_FAC_1.setBrightness(7);
    disp_FAC_2.setBrightness(7);
    disp_FAC_3.setBrightness(7);
    disp_FAC_1.setSegments(DisplayOverrideValue);
    disp_FAC_2.clear();
    disp_FAC_3.clear();
  }
  else
  {
    if (globals.battery_level < BAT_LOW_PERCENT && millis() % 10000 > 7000)
    {
      disp_FAC_1.setBrightness(0);
      disp_FAC_2.setBrightness(0);
      disp_FAC_3.setBrightness(0);

      disp_FAC_1.setSegments(sevenSeg_bat);
      disp_FAC_2.setSegments(sevenSeg_low);
      if (millis() % 3000 < 1500)
        disp_FAC_3.showNumberDec(globals.battery_level);
      else
        disp_FAC_3.showNumberDecEx(globals.loadvoltage * 100, 0x40);
    }
    else
    {
      disp_FAC_1.setBrightness(PersistenceData.activeFaction == FACTION_1 ? 7 : 0);
      disp_FAC_2.setBrightness(PersistenceData.activeFaction == FACTION_2 ? 7 : 0);
      disp_FAC_3.setBrightness(PersistenceData.activeFaction == FACTION_3 ? 7 : 0);

      disp_FAC_1.showNumberDecEx(PersistenceData.faction_1_timer / 60, Faction_1_dot, true, 4, 0);
      disp_FAC_2.showNumberDecEx(PersistenceData.faction_2_timer / 60, Faction_2_dot, true, 4, 0);
      disp_FAC_3.showNumberDecEx(PersistenceData.faction_3_timer / 60, Faction_3_dot, true, 4, 0);
    }
  }
}

void FactionTicker_callback()
{
  switch (PersistenceData.activeFaction)
  {
  case FACTION_1:
    PersistenceData.faction_1_timer++;
    Faction_1_dot = Faction_1_dot == 0x80 || Faction_1_dot == 0x00 ? 0x10 : Faction_1_dot << 1;
    Faction_2_dot = 0;
    Faction_3_dot = 0;
    break;

  case FACTION_2:
    PersistenceData.faction_2_timer++;
    Faction_2_dot = Faction_2_dot == 0x80 || Faction_2_dot == 0x00 ? 0x10 : Faction_2_dot << 1;
    Faction_1_dot = 0;
    Faction_3_dot = 0;
    break;

  case FACTION_3:
    PersistenceData.faction_3_timer++;
    Faction_3_dot = Faction_3_dot == 0x80 || Faction_3_dot == 0x00 ? 0x10 : Faction_3_dot << 1;
    Faction_1_dot = 0;
    Faction_2_dot = 0;
    break;

  default:
    break;
  }
}

void inputGetterTicker_callback()
{
  ProcessKeyCombos();

  uint8_t keysPressed = 0;

  keysPressed = +digitalRead(DIO_FAC_1_TRG) == FAC_1_TRG_PRESSED ? 1 : 0;
  keysPressed = +digitalRead(DIO_FAC_2_TRG) == FAC_2_TRG_PRESSED ? 1 : 0;
  keysPressed = +digitalRead(DIO_FAC_3_TRG) == FAC_3_TRG_PRESSED ? 1 : 0;

  if (keysPressed > 1)
  {
    Serial.println("ERROR: More than one Flag active - setting no Faction active");
    PersistenceData.activeFaction = NONE;
    return;
  }

  if (digitalRead(DIO_FAC_1_TRG) == FAC_1_TRG_PRESSED)
    PersistenceData.activeFaction = FACTION_1;

  if (digitalRead(DIO_FAC_2_TRG) == FAC_2_TRG_PRESSED)
    PersistenceData.activeFaction = FACTION_2;

  if (digitalRead(DIO_FAC_3_TRG) == FAC_3_TRG_PRESSED)
    PersistenceData.activeFaction = FACTION_3;
}

void powerMonitorTicker_callback()
{
  // loadvoltage and percentage is global, because of battery Monitoring

  const int bat_min_2s = 680;
  const int bat_max_2s = 840;
  const int bat_min_3s = 1020;
  const int bat_max_3s = 1260;

  float shuntvoltage = 0;
  float current_mA = 0;
  float busvoltage = 0;
  float power_mW = 0;
  int battery_level = 0;

  shuntvoltage = ina219.getShuntVoltage_mV();
  busvoltage = ina219.getBusVoltage_V();
  current_mA = ina219.getCurrent_mA();
  power_mW = ina219.getPower_mW();
  globals.loadvoltage = busvoltage + (shuntvoltage / 1000);
  switch (ConfigData.batteryType)
  {
  case BATTERY_LIPO_2S:
    battery_level = map(globals.loadvoltage * 100, bat_min_2s, bat_max_2s, 0, 100);
    globals.battery_level = battery_level < 0 ? 0 : battery_level;
    break;
  case BATTERY_LIPO_3S:
    battery_level = map(globals.loadvoltage * 100, bat_min_3s, bat_max_3s, 0, 100);
    globals.battery_level = battery_level < 0 ? 0 : battery_level;
    break;
  default:
    globals.battery_level = -1;
    break;
  }

  // Serial.printf("Battery Level: %d %%\n", globals.battery_level);
  // Serial.printf("Bus Voltage: %f V\n", busvoltage);
  // Serial.printf("Shunt Voltage: %f mV\n", shuntvoltage);
  // Serial.printf("Load Voltage: %f V\n", globals.loadvoltage);
  // Serial.printf("Current: %f mA\n", current_mA);
  // Serial.printf("Power: %f mW\n", power_mW);
}

void EEPROMCyclicPDS_callback()
{
  StorePersistence_EEPROM();
}

#ifdef WIFI_CLIENT
void wifiMaintainConnectionTicker_callback()
{
  static uint32_t WiFiFailCount = 0;
  const uint32_t WiFiFailMax = 20;

  if (wifiMulti.run(connectTimeoutMs) == WL_CONNECTED)
  {
    return;
  }
  else
  {
    if (WiFiFailCount < WiFiFailMax)
      WiFiFailCount++;
    else
      toggleWiFiAP(false);
  }
}
#endif

void toggleWiFiAP(boolean shutdown)
{
  if (WiFi.getMode() != WIFI_OFF && shutdown == true)
  {
    WiFi.mode(WIFI_OFF);
#ifdef WIFI_CLIENT
    WiFiMaintainConnectionTicker.stop();
#endif
  }
  else if (shutdown == false)
  {
    WiFi.mode(WIFI_AP);
    WiFi.softAPConfig(IPAddress(WIFI_AP_IP_GW), IPAddress(WIFI_AP_IP_GW), IPAddress(255, 255, 255, 0));
    WiFi.softAP(globals.DeviceName_ID, QUOTE(WIFI_AP_PASSWORD));
#ifdef WIFI_CLIENT
    WiFiMaintainConnectionTicker.stop();
#endif
  }
}

void SystemShutdown()
{
  StoreConfig_EEPROM();
  ESP.restart();
}

void SetBatteryType(batteryType_t type)
{
  if (ConfigData.batteryType != type)
  {
    ConfigData.batteryType = type;
    globals.requestEEAction = EE_CFG_SAVE;
    Serial.printf("Set Batterytype to %s\n", type == BATTERY_LIPO_2S ? "2s Lipo" : "3s LiPo");
  }
}

void ProcessKeyCombos()
{
  typedef enum
  {
    KEY_PRESSED,
    KEY_RELEASED
  } keyStatus_t;

  static keyStatus_t keyStatus_Fac1 = KEY_RELEASED;
  static uint8_t keyCount_Fac1 = 0;
  static keyStatus_t keyStatus_Fac2 = KEY_RELEASED;
  static uint8_t keyCount_Fac2 = 0;
  static keyStatus_t keyStatus_Fac3 = KEY_RELEASED;

  if (digitalRead(DIO_FAC_3_TRG) == FAC_3_TRG_PRESSED)
  {
    keyStatus_Fac3 = KEY_PRESSED;

    // Process FactionKey 1 ComboCounter
    if (digitalRead(DIO_FAC_1_TRG) == FAC_1_TRG_PRESSED && keyStatus_Fac1 == KEY_RELEASED)
    {
      keyStatus_Fac1 = KEY_PRESSED;
      keyCount_Fac1++;
    }

    if (digitalRead(DIO_FAC_1_TRG) != FAC_1_TRG_PRESSED)
      keyStatus_Fac1 = KEY_RELEASED;

    // Process FactionKey 2 ComboCounter
    if (digitalRead(DIO_FAC_2_TRG) == FAC_2_TRG_PRESSED && keyStatus_Fac2 == KEY_RELEASED)
    {
      keyStatus_Fac2 = KEY_PRESSED;
      keyCount_Fac2++;
    }

    if (digitalRead(DIO_FAC_2_TRG) != FAC_2_TRG_PRESSED)
      keyStatus_Fac2 = KEY_RELEASED;
  }

  if (digitalRead(DIO_FAC_3_TRG) != FAC_3_TRG_PRESSED && keyStatus_Fac3 == KEY_PRESSED)
  {
    Serial.printf("KeyCombo 1: %d | 2: %d\n", keyCount_Fac1, keyCount_Fac2);

    if (keyCount_Fac1 == 2 && keyCount_Fac2 == 2)
    {
      Serial.println("KeyCombo: WiFi AP ON");
      OverrideDisplay(sevenSeg_net, 5000);
      toggleWiFiAP(false);
    }

    keyCount_Fac1 = 0;
    keyCount_Fac2 = 0;
    keyStatus_Fac1 = KEY_RELEASED;
    keyStatus_Fac2 = KEY_RELEASED;
    keyStatus_Fac3 = KEY_RELEASED;
  }
}

void OverrideDisplay(const uint8_t *message, uint32_t time)
{
  DisplayOverrideFlag = millis() + time;
  DisplayOverrideValue = (uint8_t *)message;
}

uint32_t getESPChipID()
{
  uint32_t chipId;
  for (int i = 0; i < 17; i = i + 8)
  {
    chipId |= ((ESP.getEfuseMac() >> (40 - i)) & 0xff) << i;
  }
  return chipId;
}
#
void LoRaPublish_callback()
{

  StaticJsonDocument<200> doc;

  doc["lipo"] = globals.battery_level;
  doc["afact"] = PersistenceData.activeFaction;
  doc["fac1"] = PersistenceData.faction_1_timer;
  doc["fac2"] = PersistenceData.faction_2_timer;
  doc["fac3"] = PersistenceData.faction_3_timer;

  LoRa.beginPacket();
  serializeJson(doc, LoRa);
  LoRa.endPacket();

  Serial.printf("Sendet LoRa-Status Package\n");
}